import inspect
from typing import Any

from varorm.storage import BaseStorage
from varorm.fields import Field
from varorm.exceptions import *


class Model:
    class Meta:
        pass

    class _meta:
        key = None
        groups = None
        verbose_name = None
        init_load = False

    def __new__(cls, key=None):
        assert cls._storage is not None
        cls.key = key

        cls._fields = {}
        cls._preloaded_data = None

        for name, field in inspect.getmembers(cls):
            if isinstance(field, Field):
                cls._fields.update({
                    name: field
                })
        
        return super().__new__(cls)

    def __init__(self, *args, **kwargs) -> None:
        if self.get_meta().init_load:
            self._load_data()
        
    def _load_data(self):
        self._preloaded_data = self._storage.get(self._get_model_key())
            
    @classmethod
    def get_meta(cls):
        meta = cls._meta()

        for k, v in inspect.getmembers(cls.Meta):
            if not k.startswith("__"):
                setattr(meta, k, v)
        
        return meta
    
    def _get_model_key(self) -> str:
        return self.key or self.get_meta().key or self.__class__.__name__
    
    def _parse_field_data(self, __name, __value):
        return self._fields[__name].parse(__value)
    
    def _represent_field_data(self, __name, __value):
        value = self._fields[__name].parse(__value)
        return self._fields[__name].represent(value)
    
    def _get_from_storage(self, __name):
        if self._preloaded_data is None:
            value = self._storage.hget(self._get_model_key(), __name)
        else:
            try:
                value = self._preloaded_data[__name]
            except:
                raise VarDoesNotExistException

        return self._parse_field_data(__name, value)
    
    def _set_to_storage(self, __name, __value):
        self._storage.hset(self._get_model_key(), __name, self._represent_field_data(__name, __value))

    def __getattribute__(self, __name: str) -> Any:
        if __name == '_fields' or __name not in self._fields:
            return super().__getattribute__(__name)
        
        try:
            return self._get_from_storage(__name)
        except VarDoesNotExistException:
            field = self._fields[__name]
            if field._default is not None:
                return field._default
            elif field._null:
                return None
            raise VarDoesNotExistException

    def __setattr__(self, __name: str, __value: Any) -> None:
        if __name not in self._fields:
            return super().__setattr__(__name, __value)
        self._set_to_storage(__name, __value)
        if self.get_meta().init_load:
            self._load_data()
    
    def update(self, **data):
        validated_data = dict()
        for k, v in data.items():
            if k not in self._fields.keys():
                raise KeyNotFoundException(f"{k}")
            validated_data[k] = self._represent_field_data(k, v)
        self._storage.update(self._get_model_key(), validated_data)
        if self.get_meta().init_load:
            self._load_data()
            
    @classmethod
    def connect(cls, storage: BaseStorage):
        cls._storage = storage

 


import inspect
from django.conf import settings

from varorm.storage import BaseStorage
from varorm.models import Model
from varorm.dj import utils
from varorm.dj import globals


utils.configure_default_setting({
    "VARORM_STORAGE": "varorm.storage.MemoryStorage",
    "VARORM_STORAGE_CONFIG": {}
})

storage = utils.import_by_path(settings.VARORM_STORAGE)(**settings.VARORM_STORAGE_CONFIG)
assert isinstance(storage, BaseStorage)


models = dict()

for app in settings.INSTALLED_APPS:
    try:
        module = __import__(f"{app}.variables")
        module = getattr(module, 'variables')
    except:
        continue

    for i in inspect.getmembers(module, inspect.isclass):
        if issubclass(i[1], Model) and i[1] is not Model:
            i[1].connect(storage)
            if app in models:
                models[app].append(i)
            else:
                models[app] = [i]

globals.MODELS = models



from typing import Callable, Any
from datetime import date, datetime
from varorm import widgets
from varorm import exceptions


class Field:
    def __init__(
            self,
            base_type: Callable,
            verbose_name: str = None,
            choices: tuple = None,
            default: Any = None,
            null: bool = False,
        ) -> None:
        self._base_type = base_type
        self._verbose_name = verbose_name
        self._choices = choices
        self._default = default
        self._null = null

    def parse(self, value):
        if isinstance(value, str) and value == '':
            return None
        return self._base_type(value)
    
    def represent(self, value):
        return value
    
class IntegerField(Field):
    widget = widgets.IntegerWidget
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(int, *args, **kwargs)


class FloatField(Field):
    widget = widgets.FloatWidget
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(float, *args, **kwargs)


class BooleanField(Field):
    widget = widgets.BooleanWidget
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(bool, *args, **kwargs)

    def parse(self, value) -> bool:
        if isinstance(value, str) and value == '':
            return None
        return bool(int(value))

    def represent(self, value: bool):
        return int(value)


class CharField(Field):
    widget = widgets.CharWidget
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(str, *args, **kwargs)


class TextField(Field):
    widget = widgets.TextWidget

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(str, *args, **kwargs)


class DateField(Field):
    widget = widgets.DateWidget

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(date, *args, **kwargs)

    def parse(self, value) -> date:
        if isinstance(value, date):
            return value
        try:
            return date.fromisoformat(value)
        except:
            pass
        try:
            return date.fromordinal(int(value))
        except:
            pass
        
        try:
            return date.fromtimestamp(float(value))
        except:
            pass
        
        raise exceptions.UnsupportableDateFormatException
    
    def represent(self, value):
        return date.isoformat(value)
    

class DateTimeField(Field):
    widget = widgets.DateTimeWidget
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(datetime, *args, **kwargs)

    def parse(self, value) -> datetime:
        if isinstance(value, datetime):
            return value
        try:
            return datetime.fromisoformat(value)
        except:
            pass
        try:
            return datetime.fromordinal(int(value))
        except:
            pass
        try:
            return datetime.fromtimestamp(float(value))
        except:
            pass
        
        raise exceptions.UnsupportableDateTimeFormatException
    
    def represent(self, value):
        return datetime.isoformat(value)



import os
import json
import pickle
from typing import Any, Callable

from varorm.exceptions import VarDoesNotExistException


class BaseStorage:
    def hget(self, key: str, hkey: str) -> Any:
        raise NotImplementedError
    
    def hset(self, key: str, hkey: str, value: Any):
        raise NotImplementedError

    def get(self, key: str) -> dict:
        raise NotImplementedError

    def update(self, key: str, data: dict):
        raise NotImplementedError
    
    
class MemoryStorage(BaseStorage):
    def __init__(self) -> None:
        self._data = {}

    def hget(self, key: str, hkey: str) -> Any:
        try:
            return self._data[key][hkey]
        except:
            raise VarDoesNotExistException
    
    def hset(self, key: str, hkey: str, value: Any):
        if key not in self._data:
            self._data.update({
                key: {}
            })
        self._data[key].update({
            hkey: value
        })
    
    def get(self, key: str) -> dict:
        try:
            return self._data[key]
        except:
            return dict()

    def update(self, key: str, data: dict):
        if key not in self._data:
            self._data.update({
                key: {}
            })
        self._data[key].update(data)


class FileStorage(MemoryStorage):
    def __init__(
            self, 
            path: str, 
            load_function: Callable, # ex. json.load
            dump_function: Callable, # ex. json.dump
            is_binary_mode: bool = False, # For pickle = True
            save_on_set: bool = False
        ) -> None:
        super().__init__()

        self._path = path
        self._read_mode = 'rb' if is_binary_mode else 'r'
        self._write_mode = 'wb' if is_binary_mode else 'w'

        self._load_function = load_function
        self._dump_function = dump_function
        self._save_on_set = save_on_set


        if os.path.isfile(self._path):
            with open(self._path, self._read_mode) as f:
                self._data = self._load_function(f)
        
    def save(self):
        with open(self._path, self._write_mode) as f:
            self._dump_function(self._data, f)
    
    def hset(self, key: str, hkey: str, value: Any):
        super().hset(key, hkey, value)
        if self._save_on_set:
            self.save()
    
    def update(self, key: str, data: dict):
        super().hset(key, data)
        if self._save_on_set:
            self.save()


class JsonStorage(FileStorage):
    def __init__(self, path: str, save_on_set: bool = False) -> None:
        super().__init__(path, json.load, json.dump, False, save_on_set)


class PickleStorage(FileStorage):
    def __init__(self, path: str, save_on_set: bool = False) -> None:
        super().__init__(path, pickle.load, pickle.dump, True, save_on_set)


class RedisStorage(BaseStorage):
    def __init__(self, url: str = None, **kwargs) -> None:
        import redis
        if url is not None:
            connection = redis.Redis.from_url(url)
        else:
            connection = redis.Redis(**kwargs)

        connection.ping()
        self._connection = connection

    def hget(self, key: str, hkey: str) -> Any:
        if not self._connection.hexists(key, hkey):
            raise VarDoesNotExistException
        val = self._connection.hget(key, hkey)

        if isinstance(val, bytes):
            return val.decode()
        
        return val
        
    def hset(self, key: str, hkey: str, value: Any):
        return self._connection.hset(key, hkey, value)
    
    def get(self, key: str) -> dict:
        data = self._connection.hgetall(key)
        if data is None:
            return dict()
        
        return dict(map(lambda p: (p[0].decode(), p[1].decode()), data.items()))

    def update(self, key: str, data: dict):
        return self._connection.hset(key, mapping=data)
    

class MongoStorage(BaseStorage):
    def __init__(self, url: str = None, db_name: str = 'default', **kwargs) -> None:
        import pymongo
        if url is not None:
            client = pymongo.MongoClient(url)
        else:
            client = pymongo.MongoClient(**kwargs)

        self._db = client[db_name]

    def hget(self, key: str, hkey: str) -> Any:
        try:
            val = self._db[key].find_one({"key": hkey})
            assert val is not None
        except:
            raise VarDoesNotExistException

        val = val['value']
        if isinstance(val, bytes):
            return val.decode()
        
        return val
        
    def hset(self, key: str, hkey: str, value: Any):
        return self._db[key].replace_one({"key": hkey}, {"key": hkey, "value": value}, upsert=True)
    
    def get(self, key: str) -> dict:
        try:
            val = self._db[key]
            assert val is not None
        except:
            return dict()

        data = dict()
        for row in val.find():
            data[row['key']] = row['value']
        return data
    
    def update(self, key: str, data: dict):
        for hk, v in data.items():
            self.hset(key, hk, v)
    

class DjangoDBStorage(BaseStorage):
    def hget(self, key: str, hkey: str) -> Any:
        from varorm.dj.models import DBStorage
        row = DBStorage.objects.filter(key=key, hkey=hkey).first()
        if row is None:
            raise VarDoesNotExistException
        
        return row.value
    
    def hset(self, key: str, hkey: str, value: Any):
        from varorm.dj.models import DBStorage
        row = DBStorage.objects.filter(key=key, hkey=hkey).first() 
        if row is None:
            DBStorage(key=key, hkey=hkey, value=value).save()
        else:
            DBStorage.objects.filter(id=row.id).update(value=value)

    def get(self, key: str) -> dict:
        from varorm.dj.models import DBStorage
        data = dict()
        for row in DBStorage.objects.filter(key=key).all():
            data[row.hkey] = row.value
        return data

    def update(self, key: str, data: dict):
        for hk, v in data.items():
            self.hset(key, hk, v)
    


from typing import Any
from dataclasses import dataclass
from django.conf import settings
from varorm.exceptions import VarDoesNotExistException
from varorm import widgets


def import_by_path(name: str):
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def configure_default_setting(default_settings: dict):
    for k, v in default_settings.items():
        try:
            getattr(settings, k)
        except:
            setattr(settings, k, v)


def create_fake_model(name: str, models):
    class FakeMeta:
        abstract            = not True
        app_label           = __name__.split('.')[-2]
        module_name         = None
        verbose_name        = None
        verbose_name_plural = None
        get_ordered_objects = list
        swapped = False
        app_config = ''
        model_name = name
        object_name = name

        def get_add_permission(self):    return 'add_%s'    % self.module_name
        def get_change_permission(self): return 'change_%s' % self.module_name
        def get_delete_permission(self): return 'delete_%s' % self.module_name
        def __init__(self, name, verbose_name=None, verbose_plural=None):
            self.module_name         = name
            self.verbose_name        = verbose_name or name
            self.verbose_name_plural = verbose_plural or name

    class FakeModel:
        varorm_models = models
        _meta = FakeMeta(name)
    
    return [FakeModel]


@dataclass
class FieldAdminRepresintation:
    verbose_name: str
    field_key: str
    value: Any
    widget: widgets.Widget


def get_model_fields_represintation(cls): 
    self = cls()
    fields = []
    for field_key, field in self._fields.items():
        value = None
        try:
            value = getattr(self, field_key)
        except VarDoesNotExistException:
            pass
        fields.append(FieldAdminRepresintation(
            verbose_name=field._verbose_name or field_key.capitalize(),
            field_key=field_key,
            value=value,
            widget=field.widget(field_key, value)
        ))

    return fields


def has_django_model_permission(model, user):
    varorm_models = model.varorm_models
    for varorm_model in varorm_models:
        if has_varorm_model_permission(varorm_model, user):
            return True
        
    return False


def has_varorm_model_permission(model, user):
    if user.is_superuser:
        return True
    groups = model.get_meta().groups
    if groups == "__all__":
        return True
    if groups is None:
        return False
    return bool(set(groups) & set([group.name for group in user.groups.all()]))


from typing import Any


class Widget:
    input = '<input type="text" id="{id}" name="{id}" required {value_attr}>'
    def __init__(self, id: str, value: Any) -> None:
        value_attr = ""
        if value is not None:
            value_attr = f'value="{value}"'

        self.input = self.input.format(id=id, value_attr=value_attr)
    

class IntegerWidget(Widget):
    input = '<input type="number" id="{id}" name="{id}" required {value_attr}>'

class FloatWidget(Widget):
    input = '<input type="number" id="{id}" name="{id}" step="0.0000001" required {value_attr}>'

class DateWidget(Widget):
    input = '<input type="date" id="{id}" name="{id}" required {value_attr}>'

class DateTimeWidget(Widget):
    input = '<input type="datetime-local" id="{id}" name="{id}" required {value_attr}>'

class CharWidget(Widget):
    input = '<input type="text" id="{id}" name="{id}" required {value_attr}>'

class TextWidget(Widget):
    input = '<textarea id="{id}" name="{id}" required>{value}</textarea>'
    
    def __init__(self, id: str, value: Any) -> None:
        if value is None:
            value = ""
        self.input = self.input.format(id=id, value=value)


class BooleanWidget(Widget):
    input = '''<select required id="{id}" name="{id}">
                <option value="" {selected_none}>-</option>
                <option value="1" {selected_1}>True</option>
                <option value="0" {selected_0}>False</option>
               </select>'''
    def __init__(self, id: str, value: Any) -> None:
        selected_0 = ""
        selected_1 = ""
        selected_none = ""
        if value is not None:
            if value:
                selected_1 = "selected"
            else:
                selected_0 = "selected"
        else:
            selected_none = "selected"

        self.input = self.input.format(id=id, selected_0=selected_0, selected_1=selected_1, selected_none=selected_none)


# Generated by Django 4.1.5 on 2024-09-10 13:33

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='DBStorage',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=255)),
                ('hkey', models.CharField(max_length=255)),
                ('value', models.TextField(null=True)),
            ],
            options={
                'db_table': 'varorm_db_storage',
                'unique_together': {('key', 'hkey')},
            },
        ),
    ]



from django.shortcuts import render
from django.http import HttpRequest
from varorm.dj.globals import MODELS
from varorm.dj import utils

def get_app_models(request: HttpRequest, app: str):
    if request.method == 'POST':
        data = dict(request.POST)
        model_name = data['dj.model_name'][0]
        field_key = data['dj.field_key'][0]
        value = data[field_key][0]
        for mn, m in MODELS[app]:
            if mn == model_name and utils.has_varorm_model_permission(m, request.user):
                setattr(m(), field_key, value)
                break

    return render(request, "models_list.html", {
        "models": [
            (v.get_meta().verbose_name or k, utils.get_model_fields_represintation(v)) 
            for k, v in MODELS[app] 
            if utils.has_varorm_model_permission(v, request.user)
        ]
    })




from django.db import models


class DBStorage(models.Model):
    key = models.CharField(max_length=255)
    hkey = models.CharField(max_length=255)
    value = models.TextField(null=True)

    class Meta:
        unique_together = ('key', 'hkey')
        db_table = 'varorm_db_storage'



from setuptools import setup, find_packages

with open("README.md", 'r') as f:
    long_description = f.read()

with open("LICENSE", 'r') as f:
    license = f.read()

project_urls = {
  'GitHub': 'https://github.com/sidqdev/varorm',
  'Telegram': 'https://t.me/sidqdev'
}


setup(
    name='varorm',
    version='0.1.6',
    author='Sidq',
    author_email='abba.dmytro@gmail.com',
    description='Variable orm with django admin support',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    license=license,
    project_urls=project_urls,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Operating System :: OS Independent',
    ],
    package_data={'varorm.dj': ['templates/*.html']},
    include_package_data=True,
    python_requires='>=3.6',
)




class VarDoesNotExistException(Exception):
    pass

class UnsupportableDateFormatException(Exception):
    pass

class UnsupportableDateTimeFormatException(Exception):
    pass

class KeyNotFoundException(Exception):
    pass



from django.contrib import admin
from django.urls import path
from django.http import HttpRequest

from varorm.dj import views, globals, utils

    
class VarormAdmin(admin.ModelAdmin):
        def has_module_permission(self, _):
            return True
        
        def get_model_perms(self, _):
            return {"view": True}

for app, models in globals.MODELS.items():
    django_model = utils.create_fake_model(app, [x[1] for x in models])
    admin.site.register(django_model, VarormAdmin)


_build_app_dict = admin.site._build_app_dict

def custom_build_app_dict(request: HttpRequest, label=None):
    app_dict = _build_app_dict(request, label)
    if 'dj' in app_dict:
        models = app_dict['dj']['models']
        models = list(filter(lambda x: utils.has_django_model_permission(x['model'], request.user), models))
        if len(models) == 0:
            del app_dict['dj']
        else:
            app_dict['dj']['models'] = models

    return app_dict

admin.site._build_app_dict = custom_build_app_dict


get_urls = admin.site.get_urls

def custom_get_urls():
    urls = get_urls()
    dj_urls = [
        path("dj/<str:app>/", views.get_app_models),
    ]
    urls = dj_urls + urls
    return urls

admin.site.get_urls = custom_get_urls